import math
from random import random

from django.db.models import Case, Count, F, Sum, When

from shiftregister.importer.models import Event


class Bucket:
    def __init__(self, selector, sub_buckets=[]):
        if not isinstance(sub_buckets, list) or not all(
            map(lambda b: isinstance(b, Bucket), sub_buckets)
        ):
            raise ValueError("sub_buckets must be a list of instances of class Bucket")

        self.selector = selector
        self.sub_buckets = sub_buckets

    def __iter__(self):
        for bucket in self.sub_buckets:
            yield bucket.selector, bucket

    def has_sub_buckets(self):
        return len(self.sub_buckets) > 0


def distribute(total_slots, total_assignments, pool, qs=None):
    if not isinstance(pool, Bucket):
        raise ValueError("pool must be an instance of class Bucket")

    qs = (qs or Event.objects).filter(pool.selector)

    for selector, bucket in pool:
        bucket_qs = qs.filter(selector).annotate(
            real_required_helpers=Case(
                When(required_helpers=0, then=F("room__required_helpers")),
                default=F("required_helpers"),
            )
        )

        bucket_slots = bucket_qs.aggregate(sum=Sum("real_required_helpers"))["sum"]

        extra_chance, bucket_assignments = math.modf(
            bucket_slots / total_slots * total_assignments
        )
        bucket_assignments = int(bucket_assignments)
        if extra_chance > random():
            bucket_assignments += 1

        if bucket.has_sub_buckets():
            yield from distribute(bucket_slots, bucket_assignments, bucket, qs)
        else:
            yield bucket_assignments, bucket_qs

        total_slots -= bucket_slots
        total_assignments -= bucket_assignments
